package interfaceProofTree;


import javax.swing.*; // JPanel (entre otros)
import javax.swing.tree.*;
import java.awt.*; // para borderLayout
import java.util.*;
import java.awt.event.*; // para borderLayout
import javax.swing.border.*;

import proofTree.*;

public class Oracle extends JPanel implements ActionListener{
   public final static int TOP_DOWN_STRATEGY=0;
   public final static int DIVIDE_AND_QUERY_STRATEGY=1;
   public final static int COMBINED_STRATEGY=2;
   public final static int NO_STRATEGY=3;

   final int gap = 15;
   final int panelHeight = 180;
   final int frameHeightInc = 50;


   ImageIcon nonvalidIcon,validIcon,dontknowIcon,trustedIcon;
   int strategy;
   ProofTree tree;
   DDTWindow frame;
   JLabel topLabel;
   JPanel centerPanel=null; // the center panel

   // only for top down stategy
   Vector childrenNodes; // to keep the children of the selected node
   Vector buttons; // to keep the group button

   // only for divide and query strategy
   DefaultMutableTreeNode node;
   ButtonGroup group;

   // only for the combined strategy
   int lastStepStrategy;

    public Oracle(int theStrategy,
                   ImageIcon theNonvalidIcon,
                      ImageIcon theValidIcon,
                      ImageIcon theDontknowIcon,
                      ImageIcon theTrustedIcon,
		      DDTWindow theFrame,
		       ProofTree theTree) {
     strategy = theStrategy;
     tree = theTree;
     frame = theFrame;
     nonvalidIcon = theNonvalidIcon;
     validIcon    = theValidIcon;
     dontknowIcon = theDontknowIcon;
     trustedIcon  = theTrustedIcon;

     resetNavigation();
    }

    public void endNavigation(){
     Dimension d = frame.getSize();
     resetNavigation();
     repaint();
    }

    public void resetNavigation() {
     removeAll();
     setVisible(false);
    }

    public void initiateNavigation() {

     // in case there is any old component, remove them
     resetNavigation();

     // if there is no strategy we cannot proceed
     if (getStrategy() == NO_STRATEGY) {
	return;
     }

     Font fuente = new Font("Arial", Font.PLAIN, tree.getFontSize());
     setFont(fuente);
     setLayout(new BorderLayout());

     JPanel south = new JPanel();
     JButton nextButton =  new JButton("Next");
     fuente = new Font("Arial", Font.BOLD, tree.getFontSize());
     nextButton.setFont(fuente);
     nextButton.addActionListener(this);
     nextButton.setToolTipText("Proceed the navigation");
     south.add(nextButton);
     add(south,BorderLayout.SOUTH);

     topLabel = new JLabel("",SwingConstants.CENTER);
     fuente = new Font("Arial", Font.BOLD,tree.getFontSize()+4);
     topLabel.setFont(fuente);
     JPanel panelLabel = new JPanel();
     panelLabel.setLayout(new BorderLayout());
     panelLabel.add(topLabel,BorderLayout.CENTER);
     add(panelLabel,BorderLayout.NORTH);

     // move the division bar   
     frame.openOraclePanel();

     setVisible(true);
     displayQuestion();


    }
    public void displayQuestion() {

    switch(getStrategy()) {
       case TOP_DOWN_STRATEGY:
      		//resetButton.setToolTipText("Restart the navigation from the current selected node");
      		topLabel.setText("Top-Down Strategy\n");
          	topDown();
	  break;
       case DIVIDE_AND_QUERY_STRATEGY:
      		//resetButton.setToolTipText("Select the center of the tree and to proceed");
      		topLabel.setText("Divide & Query Strategy\n");
          	divideAndQuery();
	  break;
	 case COMBINED_STRATEGY:
                topLabel.setText("Combining Strategies\n");
		combined();
	};
    } // displayQuestion


 private void resetCenterPanel() {
      if (centerPanel != null)
         remove(centerPanel);
 }

 private void setNewQuestion(JPanel newPanel) {
      resetCenterPanel();
      if (newPanel != null) {
         centerPanel = newPanel;
         centerPanel.setBorder( new BevelBorder(BevelBorder.RAISED) );
         add(centerPanel,BorderLayout.CENTER);
      }
      repaint();
     } // newQuestion

    public int getStrategy() {
      return strategy;
    }

    public void setStrategy(int newStrategy) {
     if (newStrategy != strategy) {
        strategy = newStrategy;
	endNavigation();
        initiateNavigation();

      }
    }

 // the listener call the corresponding function depending on the button
 // and the current strategy
 public void actionPerformed(ActionEvent  evt) {
   int strategy = getStrategy();

     if (strategy == TOP_DOWN_STRATEGY)
	nextTopDownListener();
    else      if (strategy == DIVIDE_AND_QUERY_STRATEGY)
			nextDivideAndQueryListener();
	         else if (strategy == COMBINED_STRATEGY)
			nextCombinedListener();

 } // actionPerformed



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Top - Down
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

 public void topDown(){
     childrenNodes = null;
     buttons = null;
     resetCenterPanel();
     // obtaing the node variable
     DefaultMutableTreeNode  node = (DefaultMutableTreeNode)
                                                 tree.getLastSelectedPathComponent();
     if (node == null ) node = tree.getRoot();


     // if there is any deepest non-valid tree in this branch then select it
     tree.deepestFrom(node);
     node = (DefaultMutableTreeNode)
                                                 tree.getLastSelectedPathComponent();


     // count the number of dontknow childrens of this node
     int totalChild = node.getChildCount();
     int childNumber = numberDontKnowChildren(node);

      if (childNumber == 0 && !tree.getNewBuggy()) {
           JOptionPane.showMessageDialog(null,"The Top-Down strategy cannot proceed","DDT Information - Top-Down Strategy",JOptionPane.ERROR_MESSAGE);
	   resetNavigation();
	   return;
      }
      if (childNumber == 0)
          return;


      JPanel panel = new JPanel();

      // panel.setLayout(new GridLayout(1,2));
     GridBagLayout gridbag = new GridBagLayout();
      panel.setLayout(gridbag);
      GridBagConstraints c = new GridBagConstraints();

      JPanel leftPanel = new JPanel();
      leftPanel.setLayout(new GridLayout(1+childNumber,1));
      JPanel rightPanel = new JPanel();
      GridLayout  layout = new GridLayout(1+childNumber,4);
      layout.setHgap(gap);
      rightPanel.setLayout(layout);

      c.fill = GridBagConstraints.BOTH;
     gridbag.setConstraints(leftPanel, c);
      panel.add(leftPanel);
      c.fill = GridBagConstraints.NONE;
     c.gridwidth = GridBagConstraints.REMAINDER;
     gridbag.setConstraints(rightPanel, c);
      panel.add(rightPanel);

      Font fuente = new Font("Arial", Font.BOLD, tree.getFontSize()+3);
      JLabel basicFactsTitle = new JLabel("  Basic Facts");
      basicFactsTitle.setFont(fuente);
      leftPanel.add(basicFactsTitle);
      stateIcons(rightPanel);

      childrenNodes = new Vector();
      buttons = new Vector();
      childrenNodes.setSize(childNumber);
      buttons.setSize(childNumber);

      ButtonGroup group;
      fuente = new Font("Arial", Font.PLAIN, tree.getFontSize());

      int i=0;
      for (int j=0; j<totalChild; j++) {
       DefaultMutableTreeNode child = (DefaultMutableTreeNode) node.getChildAt(j);
       NodeInfo nodeInfo =(NodeInfo)(child.getUserObject());
       if (nodeInfo.getNodeState() == NodeInfo.DONTKNOW) {
           childrenNodes.setElementAt(child,i);
           group = stateButtons(rightPanel);
           buttons.setElementAt(group,i);
           JLabel basicFact = new JLabel("  "+ nodeInfo.toString());
           basicFact.setFont(fuente);
           leftPanel.add(basicFact);
	   i++;
	 }
      }


    setNewQuestion(panel);
    }

/////////////////////////////////////////////////////////////////////////////////
private int numberDontKnowChildren(DefaultMutableTreeNode node) {
     int  totalChildNumber = node.getChildCount();
     int childNumber = 0;

      for (int i=0; i<totalChildNumber; i++) {
       DefaultMutableTreeNode child = (DefaultMutableTreeNode) node.getChildAt(i);
       NodeInfo nodeInfo =(NodeInfo)(child.getUserObject());
         if (nodeInfo.getNodeState() == NodeInfo.DONTKNOW)
	     childNumber++;
       }	 // for
      return childNumber;
}
/////////////////////////////////////////////////////////////////////////////////

    private ButtonGroup stateButtons(JPanel panel) {
      ButtonGroup group = new ButtonGroup();
      for (int i=0; i<4; i++) {
          JRadioButton button = new JRadioButton();
	  group.add(button);
	  panel.add(button);
          if (i==2) // dontknow node
	     button.setSelected(true);
      } // for
      return group;
    } // stateButtons

    private void stateIcons(JPanel panel) {
     panel.add(new JLabel("",validIcon,SwingConstants.LEFT));
     panel.add(new JLabel("",nonvalidIcon,SwingConstants.LEFT));
     panel.add(new JLabel("",dontknowIcon,SwingConstants.LEFT));
     panel.add(new JLabel("",trustedIcon,SwingConstants.LEFT));
    } // stateButtons

void nextTopDownListener() {


   if (buttons == null) {
           JOptionPane.showMessageDialog(null,"The Top-Down strategy cannot proceed","DDT - Top-Down Strategy", JOptionPane.ERROR_MESSAGE);
	   resetNavigation();
           return;
   }

    ButtonGroup group;
    int n = buttons.size(), currentState,newState=-1;
    boolean anyNonValid = false;
    int minimumNumberDontKnow=0;
    int numberDontKnow;
    DefaultMutableTreeNode nextRoot=null;
    DefaultMutableTreeNode node,nextNode=null;
    NodeInfo nodeInfo;

    for (int i=0; i<n; i++) {
        group = (ButtonGroup) buttons.get(i);
	node  = (DefaultMutableTreeNode) childrenNodes.get(i);
        nodeInfo =(NodeInfo)(node.getUserObject());
	currentState = nodeInfo.getNodeState();
	numberDontKnow = numberDontKnowChildren(node);

	Enumeration e = group.getElements();
	int j=0;
        while (e.hasMoreElements()) {
         JRadioButton button = (JRadioButton) e.nextElement();
	 if (button.isSelected()) {
	     newState = posToState(j);
	     if (newState == NodeInfo.NONVALID) {
	          if (!anyNonValid) {
		     nextRoot = node;
		     minimumNumberDontKnow = numberDontKnow;
		  } else  // check if it is the new minimun
		     if (numberDontKnow < minimumNumberDontKnow) {
		        nextRoot = node;
		        minimumNumberDontKnow = numberDontKnow;
		     }
	          anyNonValid = true;
                  // tree.selectNode(node);
	         } // if
	   } // if
	   j++;
	} // while

        nodeInfo.setNodeState(newState);
	tree.changedState(node,currentState,newState,"(by user )(top-down strategy)");
    } // for


    if (!anyNonValid) {
         if (!tree.getNewBuggy())
           JOptionPane.showMessageDialog(null,"At least one non-valid node is neccessary  to proceed","DDT Information - Top-Down Strategy",JOptionPane.ERROR_MESSAGE);
           return;
    } else {

       // new root
	tree.setRoot(nextRoot);
	tree.selectNode(nextRoot);

	// conmtinue the navigation
	if (strategy == COMBINED_STRATEGY)
		combined();
	else
           	topDown();
    	} // else

 } // ActionPerformed


private int posToState(int pos) {
       int state=-1;
       switch(pos) {
                  case 0:
		     state = NodeInfo.VALID;
		      break;

                  case 1:
		      state = NodeInfo.NONVALID;
                      break;

		  case 2:
		      state = NodeInfo.DONTKNOW;
                      break;

                  case 3:
		      state = NodeInfo.TRUSTED;
                      break;

                  case 4:
		      state = NodeInfo.BUGGY;
                      break;
             }  // switch
  return state;
 } // posToState



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Divide and query
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    private void divideAndQuery(){
     JPanel panel = new JPanel();
     panel.setLayout(new BorderLayout());

     tree.divide();
     // obtaing the node variable
     node = (DefaultMutableTreeNode)
                                                 tree.getLastSelectedPathComponent();
       if (node == null ) node=tree.getRoot();

      GridBagLayout gridbag = new GridBagLayout();
      panel.setLayout(gridbag);
      GridBagConstraints c = new GridBagConstraints();
      //panel.setLayout(new GridLayout(1,2));
      JPanel leftPanel = new JPanel();
      leftPanel.setLayout(new GridLayout(2,1));
      JPanel rightPanel = new JPanel();
      GridLayout layout = new GridLayout(2,3);
      layout.setHgap(gap);
      rightPanel.setLayout(layout);
      c.fill = GridBagConstraints.BOTH;
     gridbag.setConstraints(leftPanel, c);
      panel.add(leftPanel);
      c.fill = GridBagConstraints.NONE;
     c.gridwidth = GridBagConstraints.REMAINDER;
     gridbag.setConstraints(rightPanel, c);
      panel.add(rightPanel);


      prepareLeftPanel(leftPanel);
      prepareRightPanel(rightPanel);

     setNewQuestion(panel);

    }




private void prepareLeftPanel(JPanel panel) {
      Font fuente = new Font("Arial", Font.BOLD, tree.getFontSize()+3);
      JLabel basicFactsTitle = new JLabel("  Basic Fact");
      basicFactsTitle.setFont(fuente);
      panel.add(basicFactsTitle);

      fuente = new Font("Arial", Font.PLAIN, tree.getFontSize());
      NodeInfo nodeInfo =(NodeInfo)(node.getUserObject());
       JLabel basicFact = new JLabel("  "+ nodeInfo.toString());
       basicFact.setFont(fuente);
       panel.add(basicFact);
} // prepareLeftPanel

private void prepareRightPanel(JPanel panel) {
      // icons
      panel.add(new JLabel("",validIcon,SwingConstants.LEFT));
      panel.add(new JLabel("",nonvalidIcon,SwingConstants.LEFT));
      panel.add(new JLabel("",trustedIcon,SwingConstants.LEFT));

       //  radio buttons
     group = new ButtonGroup();
      for (int i=0; i<3; i++) {
         JRadioButton button = new JRadioButton();
	  group.add(button);
	  panel.add(button);
	  if (i==0) button.setSelected(true);
	 }

} // prepareRightPanel

void nextDivideAndQueryListener() {
       int newState=-1;
       NodeInfo nodeInfo =(NodeInfo)(node.getUserObject());
	int currentState = nodeInfo.getNodeState();
	if (currentState != NodeInfo.DONTKNOW) {
           JOptionPane.showMessageDialog(null,"The Divide  & Query strategy cannot proceed","DDT Information",JOptionPane.ERROR_MESSAGE);
	   resetNavigation();
           return;
	 }
	Enumeration e = group.getElements();
	int j=0;
        while (e.hasMoreElements()) {
         JRadioButton button = (JRadioButton) e.nextElement();
	 if (button.isSelected()) {
	     newState = j==0 ? NodeInfo.VALID
	                                :  (j==1 ? NodeInfo.NONVALID :NodeInfo.TRUSTED);
	   } // if
	   j++;
	} // while
        nodeInfo.setNodeState(newState);
	tree.changedState(node,currentState,newState,"(by user )(divide & query strategy)");
        tree.selectNode(node);
	if (newState == NodeInfo.NONVALID)
	   tree.setRoot(node);
	else
	   tree.removeSubtree(node);

if (strategy == COMBINED_STRATEGY)
	combined();
else
  	divideAndQuery();
} // nextDivideAndQueryListener

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Combined Strategy
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private void combined() {

   lastStepStrategy = tree.selectStrategy();

 switch (lastStepStrategy) {
  case TOP_DOWN_STRATEGY:
           	topDown();
  		break;
  case DIVIDE_AND_QUERY_STRATEGY:
           	divideAndQuery();
  		break;
 }  // switch

} // combined

private void nextCombinedListener() {
 switch (lastStepStrategy) {
  case TOP_DOWN_STRATEGY:
           nextTopDownListener();
  		break;
  case DIVIDE_AND_QUERY_STRATEGY:
           nextDivideAndQueryListener();
  		break;
 }  // switch
} // nextCombinedListener

} // Oracle

