

package parserString;


import java.io.*;
import java.util.Vector;
import javax.swing.tree.*;
import java.beans.*; // XMLDecoder & XMLEncoder

import proofTree.*;

/** final methods for file input/output
    uses the NodeInfo class */
public class ProofTreeIO {


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//			l o a d
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// load a text file containing a tree.
// the root of the tree is returned
  public  static DefaultTreeModel loadText(String file) {
    File inputFile = new File(file);
    DefaultTreeModel model;
    DefaultMutableTreeNode root;

    if (inputFile.exists() && inputFile.canRead()) {
         try{
             // read the text file and transform it into a DefaultTreeModel
             model = createTree( readTreeInformation(inputFile) );
         } catch(Exception e){
                e.printStackTrace();
                root = new DefaultMutableTreeNode("Error");
                model  = new DefaultTreeModel(root);

           }
    }
    else {
      root = new DefaultMutableTreeNode(" File not found: "+inputFile.getPath());
      model  = new DefaultTreeModel(root);

     }

     return model;
  } // load




  private static NodeInfo getNextNode(StringIndex si) {
   final int firstCar ='#'; // mark of the start of a node information
   final int lastCar='#'; // mark of the end of a node information


   String inputString = si.getString(); // it is easier to deal with String
   int index = si.getIndex();
   index = inputString.indexOf(firstCar,index);
   index++;

   // extract the basic fact
   int endPos =  inputString.indexOf(lastCar,index);
   String arrow = new String(inputString.substring(index, endPos));
   index = endPos+1;

   // extract the name of the function
   endPos =  inputString.indexOf(lastCar,index);
   String name = new String(inputString.substring(index, endPos));
   index = endPos+1;

   // extract the rule number
   endPos =  inputString.indexOf(lastCar,index);
   String number = new String(inputString.substring(index, endPos));
   index = endPos+1;

   // extract the state
   endPos =  inputString.indexOf(lastCar,index);
   String stateString = new String(inputString.substring(index, endPos));
   int state = Integer.parseInt(stateString);
   index = endPos+1;

   // extract the entailed basic facts
   Vector entailed = new Vector();
   while(inputString.charAt(index) == '{') {
        index++;
   	endPos =  inputString.indexOf('}',index);
	String entailedArrow = new String(inputString.substring(index, endPos));
   	entailed.addElement( entailedArrow);
        index = endPos+1;
   } // while

   NodeInfo node = new NodeInfo(arrow,name,number,state,entailed);

   // restore the index information in the input structure
   index++;
   si.setIndex(index);
   return node;
  } // getNextNode


  // reads the whole string contained in the file.
   private static String readTreeInformation(File inputFile) {
      String inputString;
      try{
        FileInputStream input = new FileInputStream(inputFile);
        byte bt[] = new byte[(int) inputFile.length()];
        int totalBytes = input.read(bt);
        inputString = new String(bt);
        input.close();
      } catch (IOException e) {
                e.printStackTrace();
                inputString = new String("#error#");
        }

     return inputString;
   } // readTreeInformation



   // creates the JTree component from the input string
  private static  DefaultTreeModel createTree(String inputString) {

   DefaultMutableTreeNode root,currentNode;
   DefaultTreeModel model;

   NodeInfo nodeInfo;
   int index=0;
   StringIndex si=null; // the string and the index together

   si = new StringIndex(inputString,index);
   nodeInfo = getNextNode(si);
   index = si.getIndex();
   root = new DefaultMutableTreeNode(nodeInfo);

   // the first node of the model
   model  = new DefaultTreeModel(root);

   currentNode = root;

   // while there is some node not included in the tree ...
   while (inputString.indexOf("#",index)!=-1) {

     // the next node can be either:
     //         - The brother of the current node
     //         - The next child of the current node
     char c = inputString.charAt(index);
     if (c == '$') // brother
       {
        currentNode = (DefaultMutableTreeNode) currentNode.getParent();
        // in the next loop the node will be the next child of the parent
        index++;
        }
     else {
     // get the next node
     si = new StringIndex(inputString,index);
     String basicfact = inputString.substring(index);
     nodeInfo = getNextNode(si);
     DefaultMutableTreeNode node = new DefaultMutableTreeNode(nodeInfo);
     index =si.getIndex();

     // it is always the last child of the list
     model.insertNodeInto(node,currentNode,currentNode.getChildCount());

     // the inserted node is the new current node
     currentNode = node;
     }

   } // while

   return model;


  } // createTree

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  // creates the JTree component from the input string
  private static  DefaultTreeModel createTree2(String inputString) {

   DefaultMutableTreeNode root,currentNode;
   DefaultTreeModel model;

   NodeInfo nodeInfo;
   int index=0;
   StringIndex si=null; // the string and the index together

   si = new StringIndex(inputString,index);
   nodeInfo = getNextNode(si);
   index = si.getIndex();
   root = new DefaultMutableTreeNode(nodeInfo);

   // the first node of the model
   model  = new DefaultTreeModel(root);

   currentNode = root;

   // while there is some node not included in the tree ...
   while (inputString.indexOf("#",index)!=-1) {

     // the next node can be either:
     //         - The brother of the current node
     //         - The next child of the current node
     char c = inputString.charAt(index);
     if (c == '$') // brother
       {
        currentNode = (DefaultMutableTreeNode) currentNode.getParent();
        // in the next loop the node will be the next child of the parent
        index++;
        }
     else {
     // get the next node
     si = new StringIndex(inputString,index);
     nodeInfo = getNextNode(si);
     DefaultMutableTreeNode node = new DefaultMutableTreeNode(nodeInfo);
     index =si.getIndex();

     // it is always the last child of the list
     model.insertNodeInto(node,currentNode,currentNode.getChildCount());

     // the inserted node is the new current node
     currentNode = node;
     }

   } // while

   return model;


  } // createTree2

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public static void saveXml(String fileName,ProofTree tree) {
 try{
   	XMLEncoder e = new XMLEncoder (
   					new BufferedOutputStream(
							new FileOutputStream(fileName)));
         DefaultMutableTreeNode node = (DefaultMutableTreeNode)tree.getRoot();
   	e.writeObject(node);
   	e.close();
    } catch(Exception e){
                e.printStackTrace();
                        System.out.println("Error escribiendo el fichero ...");

     }
} // saveXml

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public static void loadXml(String file, ProofTree tree) {
 DefaultMutableTreeNode node = null;

 try{
   	XMLDecoder d = new XMLDecoder (
   					new BufferedInputStream(
							new FileInputStream(file)));
   	node = (DefaultMutableTreeNode) d.readObject();
	tree.setRoot(node);
   	d.close();
    } catch(Exception e){
                e.printStackTrace();
                        System.out.println("Error leyendo el fichero ...");

     }

 } // loadXml

} // ProofTreeIO


