
package parserString;
/*
 * Created on 25-sep-2005
 */

/**
 * @author Rafael Caballero
 * This function is convenient for dealing with strings
 */
//a string  with a index included
public class StringIndex {
  private String s;
  private int i;
  
  public StringIndex(String theString, int theIndex) {
     setString(theString); setIndex(theIndex);
  }

  public void setString(String s) { this.s = s;}
  public void incIndex() { i++;}
  public void decIndex() { i--;}
   public void setIndex(int i ) { this.i = i;}
  public String getString() { return  s;}
  public int getIndex() { return i;}
  public void skipBlanks() {
  	while (validIndex() && blank(s.charAt(i))) i++;
  }
  
  /**
   * Looks for the first occurrence of a character in the string starting
   * at the current index position. Ocurrences of the character inside
   * string or character literals are not considered. The current index position
   * is assumed to be out of any string or character literal.
   * @param c Character to look for 
   * @return position i of the first ocurrence of c such that i>=getIndex.
   *         if such does no exists the method returns -1.
   *         The position of the index is not changed
   */
  public int next(char c) {
  	int iCopy = i; // keep a copy of index 
  	// 
  	while (validIndex() && c!=getChar()) {
  		 if (getChar()=='"')  
  		 	skipString();
  		 else if (getChar()=='\'')
  		 	 skipChar();
  		 incIndex();
  	}	 
  	int result = i;	 	
  	i = iCopy;	// restore i 	
  		 	
  	return result;
  }
  
  /**
   * Looks for the next comma after the current index possition.
   * This comma must be out of any character or string literal,
   * any round bracket or list.
   * @return the index of the next comma, or a non-valid index if 
   * such comma does not exist. The index is not modified.
   */
  public int nextComma() {
  	int result = getIndex();
  	int pos    = i;
  	char c;
  	while (validIndex() && getChar()!=',') {
  		c = getChar();
  		if (c=='(') {
  			pos = matchSymbol(')');
  			i = pos;
  		} else if (c=='[') {
  			pos = matchSymbol(']');
  			i = pos;
  		} else if (c=='"') {
  			skipString();
  			incIndex();
  		} else if (c=='\'') {
  			skipChar();
  			incIndex();
  		} else incIndex();	
  	}
  	
    result = i;

  	return result;
  }
  
  
  private boolean blank(char c) {
  	return c==' ';
  }
  	
  public String subString(int start, int end) {
  	String part = s.substring(start,end);
  	 return part;
  }
  
  public String subString(int start) {
 	 return s.substring(start);
  }
  
  /**
   * The current position is assumed to be in a " character.
   * This method looks for the end of the string, skipping scape characters 
   * The index position will point to the closing " after the method call.
   */
  public void skipString() {
    incIndex();
    while (getChar()!='\"')    
    	 if (getChar() == '\\') { // rafa 20-12-05: != changed by ==
    	 	incIndex();
    	 	incIndex();
    	 } else incIndex();   	 	       	     
  }

  /**
   * The current position is assumed to be in a ' character.
   * This method looks for the end of the char literal, skipping scape characters 
   * The index position will point to the closing ' after the method call.
   */
  public void skipChar() {
    incIndex();
    while (getChar()!='\'') 
    	 if (getChar() == '\\') { // rafa 20-12-05: != changed by ==
    	 	incIndex();
    	 	incIndex();
    	 } else incIndex();
    	    	 	  
  }

  /** 
   * 
   * @return true if the index corresponds to a valid position of the string
   *         false otherwise
   */
  public boolean validIndex() {
    return validPosition(i);	
  }

 /**
  * Looks for the corresponding close char for the char whose is currently
  * at the index position. Nested values are taken into account-
  * @param close
  * @return 1+the position where the first open character is matched by
  * its corresponding close value. Undefined if the close value doesn't exist.
  * The index position is not modified.
  */
  public int matchSymbol(char close) {
  	char open = getChar();
  	int counter = 1; // the matching close will be found when counter becomes 0
  	int pos = i+1;
  	char c;
  	while(/*validPosition(pos) &&*/ counter>0){
  		c = s.charAt(pos);
  		if (c=='"') {
  			int copy = i;
  			i = pos;
  			skipString();
  			pos = i;
  			i = copy;
  		} else if (c=='\'') {
  			int copy = i;
  			i = pos;
  			skipString();
  			pos = i;
  			i = copy;
  		} 
  		else if (c==open) counter++;
  		else if (c==close) counter--;
  		if (counter>0) pos++;
  	}
  	
  	return pos;
  }
  
  public boolean validPosition(int i) {
    return i<s.length() && i>=0;	
  }

  public char getChar() { return s.charAt(i); }
  
  public String toString() {
    return getString()+" "+getIndex();
  }
} // stringIndex
