package proofTree;

import java.util.*; // para Vector

import terms.BasicFact;


  // information contained in a node
  public class NodeInfo {
        public final static int DONTKNOW=0;
        public final static int VALID=1;
        public final static int NONVALID=2;
        public final static int TRUSTED=3;
        public final static int BUGGY=4;
        
       
        protected String functionNumber;
        protected String functionInstance;
        protected Vector entails;
        protected int nodeState;
        protected int backupState;
        protected String functionName;
        protected boolean mark;
        protected BasicFact basicFact;
        
  public NodeInfo() {}
    
  public NodeInfo(String theArrow, String theRuleName, String theNumber,
	                             int theState, Vector theEntailed) {
  		setNodeState(DONTKNOW);
  		setMark(false);
  		

  		if (theArrow.charAt(0)=='$') {
                setBasicFact(null);
                setFunctionName("main");
                setFunctionNumber ("1");
                setFunctionInstance("");
                setNodeState(NONVALID);
                setEntails(null);
        }
        else  {
      			BasicFact bf = new BasicFact(theArrow);
      			setBasicFact(bf);
                setFunctionName(theRuleName);
                setFunctionNumber(theNumber);
                setNodeState(theState);
                setFunctionInstance("");
                setEntails(theEntailed);
        }
    }

    public boolean isRoot() {
      return basicFact==null;	
    }
    
    public String toString() {
    	if (basicFact==null)
            return "root";
    	else return basicFact.toString();
    }

	
    public int getNodeState() {
       return nodeState;
    }
    
    public void setNodeState(int newValue) {
	   // the root only can be nonvalid
	   if (basicFact == null && newValue != NONVALID)
	       return;
	   backupState = nodeState;
       nodeState = newValue;
    }

	// restore the backup value
	public void undoState(){
           nodeState = backupState;
	}

	// mark-unmark the node
	public void setMark(boolean value) {
	   mark = value;
	}

	public boolean getMark() {
          return mark;
	}

	public static String getStateString(int state) {
	       String s;
               switch(state) {
                  case DONTKNOW:
		      s= new String("Unknown");
                      break;

                  case VALID:
		      s= new String("Valid");
                      break;

                  case NONVALID:
		      s= new String("Non Valid");
                      break;

                  case TRUSTED:
                      s= new String("Trusted");
                      break;

                  case BUGGY:
                      s= new String("Buggy");
                      break;

		   default:
		      s= new String("Error");
             }  // switch

	     return s;
	} // getStateString

    public String getFunctionName() {
       return functionName;
    }
    
    public void setFunctionName(String newName) {
            functionName = newName;
    }

    public String getFunctionNumber() {
           return functionNumber;
    }

    public void setFunctionNumber(String newValue) {
            functionNumber = newValue;
    }
 
    public String getFunctionInstance() {
           return functionInstance;
    }
 
    public void setFunctionInstance(String newValue) {
           functionInstance = newValue;    
    }

	public Vector getEntails() {	 	
           return entails;    
	}

	public void setEntails(Vector newValue) {
           entails = newValue;
    }

	public BasicFact getBasicFact() {return basicFact; }
	public void setBasicFact(BasicFact bf) {basicFact = bf; }

    // used when a node is selected to change its state
    public int nextState() {
               switch(nodeState) {
                  case DONTKNOW:
                      nodeState = VALID;
                      break;

                  case VALID:
                      nodeState = NONVALID;
                      break;

                  case NONVALID:
                      nodeState = TRUSTED;
                      break;

                  case TRUSTED:
                      nodeState = DONTKNOW;
                      break;
             }  // switch
	   return nodeState;
        }  // nextState

    } // NodeInfo

