

// the proof tree

// uses ProofTreeIO, NodeInfo,StatusPanel

package proofTree;

import java.awt.*; // para borderLayout
import javax.swing.*; // JPanel (entre otros)
import javax.swing.tree.*;
import java.util.*;

import parserString.ProofTreeIO;
import interfaceProofTree.*;

public class ProofTree extends  JTree {

   int nNonValid, nBuggy, nValid, nDontKnow, nTrusted;
   DefaultTreeModel model;
   DefaultMutableTreeNode root;
   StatusPanel status;
   // info about the last action
   String lastAction= new String();
   int numberLastAction=0;
   // to check if there is a new buggy after the last recount
   boolean newBuggy= false;
   // name of the current loaded file
   String fileName;
   //  the main frame
   DDTWindow frame;
   // the path
   String path;
   // entailment information
   Entailment entailment;

   private int fontSize = 16;

   public ProofTree(DDTWindow theFrame,String thePath, String theFile, String fileEntailment,
                              StatusPanel theStatus){
    status = theStatus;
    frame = theFrame;
    path = thePath;

    // load the file
    onLoad(theFile);

    // load the entailment information
    entailment = new Entailment(fileEntailment+".txt");

    // listen for mouse clicks
    Click ml = new  Click(this);
    addMouseListener(ml);

     setRowHeight(20);
     Font fuente = new Font("Arial", Font.BOLD, getFontSize());
     setFont(fuente);
     setLargeModel(true);

  } // ProofTree()

   
   /**     
    * returns the font size of the program 
    * @return The font size of the navigator
    */
   public int getFontSize() { return fontSize; }

   
   // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// repaint the proof tree and also the frame
/*public void repaint(){
 if (frame != null) frame.repaint();
 super.repaint();
}*/

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// load the tree from the text file
public void onLoad(String file) {

    // if the file name already contains the extension remove it
    int n = file.indexOf(".txt");
    if  (n !=-1)
         file = file.substring(0,n);


    setFileName(file+".xml");
    model = ProofTreeIO.loadText(file+".txt");

	setModel(model);
    	root = (DefaultMutableTreeNode) model.getRoot();


    // only one node selected at each time
    getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);
    //Enable tool tips.
    ToolTipManager.sharedInstance().registerComponent(this);


    /* Make tree ask for the height of each row. */
    setRowHeight(-1);
    // the root is the auxiliar function main and is always erroneous
    expandRow(0);
    setSelectionRow(0);

    // the tree is saved in .xml format. All the file operations inside the java
    // aplications will require this format
    save(fileName);

    // set the initial values of counters
    calculate();
    repaint();

}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public void save(String file) {
 ProofTreeIO.saveXml(file,this);
 setFileName(file);
}

public void open(String file) {
  //removeSubtree(getRoot());
 ProofTreeIO.loadXml(file,this);
  setFileName(file);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// returns the current name of the loaded file
public String getFileName() {
   return fileName;
}

// changes the name of the current filename and displays the new name
// it does not load actually the file
public void setFileName(String newFileName) {
   fileName = newFileName;
    frame.setTitle("DDT  2.1.1 - "+fileName+" -");
}

public String getPath() {
   return path;
}


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public void initializeLastAction() {
   lastAction= new String();
   numberLastAction=0;
} // initializeLastAction

public int getLastActionNumber() {
  return numberLastAction;
}

public String getLastAction() {
  return lastAction;
}

 void reportAction(DefaultMutableTreeNode node,int pre,int post,String comment) {
    String action = " "+node.toString()+"    changed from ''"+NodeInfo.getStateString(pre)+
                                  "'' to ''"+NodeInfo.getStateString(post)+"'' "+comment+".\n";
    numberLastAction++;
    lastAction += action;
 } // reportAction


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


 // calculate the total number of nodes
 public void calculate() {
   Enumeration e = root.depthFirstEnumeration();
   newBuggy = false; // by now


   nDontKnow = nNonValid = nValid  = nBuggy = nTrusted = 0;

   while (e.hasMoreElements()) {
      DefaultMutableTreeNode node = (DefaultMutableTreeNode)e.nextElement();
      NodeInfo nodeInfo = (NodeInfo)(node.getUserObject());

      int state = nodeInfo.getNodeState();
      boolean   buggy = checkBuggy(node);


     if (buggy && state==NodeInfo.NONVALID) {
	   showBuggy(node);
   	   reportAction(node,state,NodeInfo.BUGGY,"");
	   newBuggy = true;
	 }

      switch(nodeInfo.getNodeState()) {
          case NodeInfo.DONTKNOW:
               nDontKnow++;
               break;

          case NodeInfo.VALID:
               nValid++;
               break;

          case NodeInfo.NONVALID:
               nNonValid++;
               break;

          case NodeInfo.TRUSTED:
               nTrusted++;
               break;

          case NodeInfo.BUGGY:
               nBuggy++;
               break;

      } // switch
   } // while

   status.setValues(nValid,nNonValid,nDontKnow, nBuggy, nTrusted);


   }

  public boolean getNewBuggy() {
    return newBuggy;
  }

  public DefaultMutableTreeNode getRoot(){
      return root;
   }

 public void setRoot(DefaultMutableTreeNode newRoot){
    if (model==null) return;

   removeSubtree(newRoot);
//     DefaultMutableTreeNode newRootCopy = (DefaultMutableTreeNode) newRoot.clone();
    TreePath path = getPathForRow(0);
    model.setRoot(newRoot);
    root = (DefaultMutableTreeNode) model.getRoot();
    model.valueForPathChanged(path,newRoot);


    // only one node selected at each time
    getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);

    setRowHeight(-1);

    expandRow(0);
    setSelectionRow(0);
       updatePaint();

   }

   public void removeSubtree(DefaultMutableTreeNode node){
     if (model != null)
      try {
       model.removeNodeFromParent(node);
      } catch(IllegalArgumentException e) {}
   }

   public int getNDontKnow(){
      return nDontKnow;
   }

   public int getNValid(){
      return nValid;
   }

   public int getNNonValid(){
      return nValid;
   }

   public int getNTrusted(){
      return nTrusted;
   }

   public int getNBuggy(){
      return nBuggy;
   }

   public int getNTotal(){
      return getNDontKnow()+getNValid()+getNNonValid()+getNTrusted()+getNBuggy();
   }


  // returns the number of nodes in the subtree whose root is 'node'
  public int getNTotalSubtree(DefaultMutableTreeNode node){
     int total = 0;
     Enumeration e = node.breadthFirstEnumeration();
     while (e.hasMoreElements()) {
          e.nextElement();
	  total++;
      } // while

      return total;
   } //  getNTotalSubtree

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   public int maxDepth() {
     return maxDepthFrom(getRoot());
   } // maxDepth

   // maximum depth of a subtree
   public int maxDepthFrom(DefaultMutableTreeNode node) {
     Enumeration e = node.breadthFirstEnumeration();
     int depth,maxDepth = -1;
     TreePath path = null;
     DefaultMutableTreeNode element = null;

     while (e.hasMoreElements()) {
         element = (DefaultMutableTreeNode)e.nextElement();
	 depth = element.getLevel();
	 if (depth > maxDepth)
	          maxDepth = depth;
      } // while

      return maxDepth;
   } // maxDepthFrom

    public int maxBreadth() {
     return maxBreadthFrom(getRoot());
   } // maxBreadth


   // maximum number of children in a given subtree
   public int maxBreadthFrom(DefaultMutableTreeNode node) {
     Enumeration e = node.breadthFirstEnumeration();
     int nChildren,maxChildren = -1;
     TreePath path = null;
     DefaultMutableTreeNode element = null;

     while (e.hasMoreElements()) {
         element = (DefaultMutableTreeNode)e.nextElement();
	 nChildren = element.getChildCount();
	 if (nChildren > maxChildren)
	          maxChildren = nChildren;
      } // while
     return maxChildren;
   } // maxDepthFrom

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   public void deepest() {
     deepestFrom(getRoot());
   } // deepest

   // selects the deepest non-valid node in the subtree rooted by 'node'
   // if any non-valid node can be found the 'node' itself is selected
   public void deepestFrom(DefaultMutableTreeNode node) {
     Enumeration e = node.breadthFirstEnumeration();
     int depth,maxDepth = -1;
     TreePath path = null;
     DefaultMutableTreeNode deepestNode = node;
     DefaultMutableTreeNode element = null;
     NodeInfo elementInfo;

     while (e.hasMoreElements()) {
         element = (DefaultMutableTreeNode)e.nextElement();
         elementInfo = (NodeInfo)(element.getUserObject());
         if (elementInfo.getNodeState() == NodeInfo.NONVALID) {
	       depth = element.getLevel();
	       if (depth > maxDepth) {
	          maxDepth = depth;
		  deepestNode = element;
	       }
	 }
      } // while
       // establish the new selected node
       setSelectionPath( new TreePath(deepestNode.getPath()));
   } // deepest

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


   public void deepestSameRule(DefaultMutableTreeNode node) {

     DefaultMutableTreeNode element = node;
     NodeInfo elementInfo;
     // name of the function and number of the rule
     String fName,fNumber;

     // information about the node
     elementInfo = (NodeInfo)(element.getUserObject());
     fName = elementInfo.getFunctionName();
     fNumber = elementInfo.getFunctionNumber();

    // look for the deepest descendant using the same function and program rule
    DefaultMutableTreeNode child;
    boolean descFound;
    do {
    descFound = false;
     for (int i=0; i<element.getChildCount(); i++) {
		   child = (DefaultMutableTreeNode) element.getChildAt(i);
                   elementInfo = (NodeInfo)(child.getUserObject());
		   // check whether the child uses the same function and rule
		   if (elementInfo.getFunctionName().equals(fName) &&
                       elementInfo.getFunctionNumber().equals(fNumber)) {
		       descFound = true;
                       element = child;
		    }
           }
    } while (descFound);

     TreePath path =  new TreePath(element.getPath());

     // establish the new selected node
     selectNode(element);
     updatePaint();

   } // deepest

   // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   public void divide() {

	 // obtain the path to the center starting at the root
        DefaultMutableTreeNode center = divide(getRoot());
	TreePath	  path =  new TreePath(center.getPath());

       	// establish the center as new selected node
       	setSelectionPath(path);
       	setExpandedState(path,true);
     	updatePaint();
   } // divide

   // returns the 'center' of the subtree whose subtree is 'node'
   private DefaultMutableTreeNode divide(DefaultMutableTreeNode node) {
     int totalNodes=0;
     int diference;
    DefaultMutableTreeNode selectedNode=node;
    DefaultMutableTreeNode element=null;
    DefaultMutableTreeNode subElement  = null;
    NodeInfo elementInfo = null;
    NodeInfo subElementInfo = null;

     Enumeration e = node.breadthFirstEnumeration();
     Enumeration e2;

     while (e.hasMoreElements()) {
         element = (DefaultMutableTreeNode)e.nextElement();
         elementInfo = (NodeInfo)(element.getUserObject());
         if (elementInfo.getNodeState() == NodeInfo.DONTKNOW)
            totalNodes++;
      } // while
     diference = totalNodes;

      e = node.breadthFirstEnumeration();
     while (e.hasMoreElements()) {
         element = (DefaultMutableTreeNode)e.nextElement();
         elementInfo = (NodeInfo)(element.getUserObject());

	  if (elementInfo.getNodeState() == NodeInfo.DONTKNOW) {
              // count the number of nodes in its subtree
              e2 = element.breadthFirstEnumeration();
	      int subtree = -1;
              while (e2.hasMoreElements()) {
                  subElement = (DefaultMutableTreeNode)e2.nextElement();
                  subElementInfo = (NodeInfo)(subElement.getUserObject());
		  if (subElementInfo.getNodeState() == NodeInfo.DONTKNOW)
	             subtree++;
		 } // while

	      int out = totalNodes-subtree;
	      int newDiference = out > subtree ?  out-subtree : subtree-out;

   	      if (newDiference <= diference) {
	          diference = newDiference;
		  selectedNode =  element;
	       }
	    } // if
	 } // while

	 return selectedNode;
   } // divide/1

  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  public void selectNode(DefaultMutableTreeNode node) {
     TreePath  path =  new TreePath(node.getPath());
       // establish the new selected node
       clearSelection();
       setSelectionPath(path);
       setExpandedState(path,true);
  }


      // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   // returns a suitable strategy using a number of test-nodes.
   // Each test-node in turn is considered 'buggy' and the number
   // of steps required with the two strategies is computed.
   // The strategy with the least number of steps required at the end is considered the best
   // choice
  public int selectStrategy() {


//System.out.println("Selecting strategy");

       // obtaing the selected node; or the root is there is no selected node
      DefaultMutableTreeNode node = (DefaultMutableTreeNode)  getLastSelectedPathComponent();
       if (node == null ) node = getRoot();

       // count the number of nodes in the subtree
       int numberNodes = getNTotalSubtree(node);

       int posNodes[] = indexesForTheTest(numberNodes);
/*
for (int j=0;j<posNodes.length;j++) {
     System.out.println("posNodes["+j+"] ="+posNodes[j]);
}
*/
       // perform the test with each selected node
	 int indexEnumeration=0; // the position of the current node in the enumeration
	 int indexArray=0; // the index of the node in the array we are considering
	 // data of the current node
	 DefaultMutableTreeNode element = null;
	 NodeInfo elementInfo = null;
	// valor a devolver
	int returnValue = Oracle.NO_STRATEGY;

	// topDown and Divide and query statistics
	Point totalTopDown = new Point(0,0); // x number of nodes; y number of steps
	int totalDivideAndQuery = 0;
       // values for any partiuclar node
       Point topDownData = null;
       int divideAndQueryData = 0;

//System.out.println("Perfoming the Tests");

	// traversing the tree
        Enumeration e = node.breadthFirstEnumeration();

	 // loop looking for the selected nodes
	  while (e.hasMoreElements() && indexArray<posNodes.length) {
   		element = (DefaultMutableTreeNode)e.nextElement();
		if (indexEnumeration == posNodes[indexArray]) {
		   // this node is selected
         	  elementInfo = (NodeInfo)(element.getUserObject());

		//System.out.println("Top Down Test");
		  // top-down
		  topDownData = topDownTest(node,element);
		  totalTopDown.x += topDownData.x;
		  totalTopDown.y += topDownData.y;

		//System.out.println("Divide And Query Test");
		  // divide_and_query
		  divideAndQueryData = divideAndQueryTest(node,element);
		  totalDivideAndQuery += divideAndQueryData;
/*
		  System.out.println(elementInfo);
	  	 System.out.println("         steps top-down: "+topDownData.y+"    nodes top-down: "+topDownData.x);
	 	 System.out.println("        steps divide and query: "+divideAndQueryData);
*/
		  // prepare the index of the next selected value in the array
		  indexArray++;
		}
		// prepare the index of the next element index in the enumeration
		indexEnumeration++;
	  } // while
/*
         System.out.println("Number of tested nodes: "+indexArray+" of ("+posNodes.length+")");
	 System.out.println("Total steps top-down: "+totalTopDown.y+" Total nodes top-down: "+totalTopDown.x);
	 System.out.println("Total steps divide and query: "+totalDivideAndQuery);
*/
	 returnValue =  totalTopDown.x/1.8 < totalDivideAndQuery
	                                          ? Oracle.TOP_DOWN_STRATEGY
	                                    	   : Oracle.DIVIDE_AND_QUERY_STRATEGY;

        return returnValue;
  } // selectStrategy

  private int [] indexesForTheTest(int nNodes) {
   final int max = 400;
   int n= nNodes > max ? max : nNodes;
   int m[] = new int [n]; // array that will contain the index of the test-nodes
   int i =0;
   int value;

   if (n==nNodes) // all the nodes are test nodes
     for (i=0; i<m.length;i++) m[i] = i;
   else  // select randomly the nodes
      for (i=0;i<m.length;i++) {
            boolean newValue;

	    // generate a new Value
	     do {
             value = (int) (Math.random()*nNodes);
	       newValue = true;
	       for (int j=0;j<i && newValue; j++)
	            if (m[j] == value) newValue =false;
	     } while(!newValue);



	     // insert the new element
	     for (int j=0; j<i;j++)
	          if (value<m[j]) {
		      int aux = m[j];
		      m[j] = value;
		      value = aux;
		  }
	      m[i] = value;
	     } // for

    return m;
  } // indexesFortheTest

  private Point topDownTest(DefaultMutableTreeNode node, DefaultMutableTreeNode buggy) {
        // returning value
     	Point result = new Point(0,0);

        // path to the buggy node
  	TreePath	  pathBuggy =  new TreePath(buggy.getPath());
	TreePath       pathNode = null; // for each intermediate element

	// each of the elements from node to buggy
       DefaultMutableTreeNode element = node;

       // simulate the navigation from the root to the buggy node
       while (element != buggy) {

        	// one step more
        	int childCount = model.getChildCount(element);
		result.x += childCount;
		result.y++;

		// look for the child in the path
		boolean found = false;
		for (int i=0; !found && i<childCount; i++) {
             		DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(element,i);
			pathNode =  new TreePath(child.getPath());
			if (pathNode.isDescendant(pathBuggy)) {
			   found = true;
			   element = child;
			   }
		} // for
		// only for debugging. This message should't appear
		if (!found)
		   System.out.println("Error: no child in the path found. Top-Down Test.");
            } // while

    return result;
  }

  private int divideAndQueryTest(DefaultMutableTreeNode node, DefaultMutableTreeNode buggy){
   	int result=0;
  	TreePath	 pathBuggy =  new TreePath(buggy.getPath());
	TreePath       pathCenter = null; // for each intermediate element
	DefaultMutableTreeNode element = node;
         NodeInfo elementInfo = null;

	// unmark all the nodes in the subtree
	markSubtree(node,false);

	// simultate the navigation recursively
   	while (countDontKnow(element)>1) {
       		DefaultMutableTreeNode center = divide(element);
		pathCenter = new TreePath(center.getPath());
		if (pathCenter.isDescendant(pathBuggy)) {
  			element = center; // new root for the simulation
		   	elementInfo = (NodeInfo)(element.getUserObject());
			if (!elementInfo.getMark()) {
		   		elementInfo.setMark(true);
		   		elementInfo.setNodeState(NodeInfo.NONVALID);
			} else
			    System.out.println("Error: marked center (divideAndQueryTest)");

		 }
		 else {
		   // we mark and turn valid all the nodes in the subtree whose root is center
		    validSubtree(center);
		  }
		 result++;
       	} // while

       undoMarkedSubtree(node);

     return result;
  }

  private int countDontKnow(DefaultMutableTreeNode node) {
  	int n=0;
        DefaultMutableTreeNode element;
         NodeInfo elementInfo = null;
	// traversing the tree
        Enumeration e = node.breadthFirstEnumeration();

	  while (e.hasMoreElements()) {
   		element = (DefaultMutableTreeNode)e.nextElement();
         	elementInfo = (NodeInfo)(element.getUserObject());
		if (elementInfo.getNodeState()==NodeInfo.DONTKNOW)  n++;
         } // while
	return n;
  } // countDontKnow

  private void validSubtree(DefaultMutableTreeNode node) {
        DefaultMutableTreeNode element;
         NodeInfo elementInfo = null;
	// traversing the tree
        Enumeration e = node.breadthFirstEnumeration();

	 // mark all the elements
	  while (e.hasMoreElements()) {
   		element = (DefaultMutableTreeNode)e.nextElement();
         	elementInfo = (NodeInfo)(element.getUserObject());
		if (!elementInfo.getMark()) {
		   elementInfo.setMark(true);
		   elementInfo.setNodeState(NodeInfo.VALID);
		   }
         } // while
    } // validSubtree

    // restore the state of all the marked nodes
    private void undoMarkedSubtree(DefaultMutableTreeNode node) {
        DefaultMutableTreeNode element;
         NodeInfo elementInfo = null;
	// traversing the tree
        Enumeration e = node.breadthFirstEnumeration();

	 // mark all the elements
	  while (e.hasMoreElements()) {
   		element = (DefaultMutableTreeNode)e.nextElement();
         	elementInfo = (NodeInfo)(element.getUserObject());
		if (elementInfo.getMark()) {
		   elementInfo.setMark(false);
		   elementInfo.undoState();
		   }
         } // while
    } // undoMarkedSubtree


    // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    public void markSubtree(DefaultMutableTreeNode node, boolean mark) {
        DefaultMutableTreeNode element;
         NodeInfo elementInfo = null;
	// traversing the tree
        Enumeration e = node.breadthFirstEnumeration();

	 // mark all the elements
	  while (e.hasMoreElements()) {
   		element = (DefaultMutableTreeNode)e.nextElement();
         	elementInfo = (NodeInfo)(element.getUserObject());
		elementInfo.setMark(mark);
         }
    }

    // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  public void removeNodes(boolean trusted, boolean valid) {
     Enumeration e = getRoot().breadthFirstEnumeration();
  	 while ( e.hasMoreElements()) {
  	 	 // next element
         DefaultMutableTreeNode element = (DefaultMutableTreeNode)e.nextElement();
         NodeInfo elementInfo = (NodeInfo)(element.getUserObject());
         int elementState = elementInfo.getNodeState();
         
         // if it must be removed
         if ( (trusted && elementState == NodeInfo.TRUSTED) ||
	       (valid && elementState == NodeInfo.VALID) ) {
	       
           // we must eliminate the node but save the children
           DefaultMutableTreeNode parent = (DefaultMutableTreeNode) element.getParent();
           DefaultMutableTreeNode child;
           int numChildren = element.getChildCount();
           Vector childrenV = new Vector();
		   for (int i=0; i<numChildren; i++) {
		  		child = (DefaultMutableTreeNode) element.getChildAt(i);
		  		childrenV.addElement(child);
		   }
		   
		   
	     model.removeNodeFromParent(element);
	 
	     // now add the children to the parent
         int pos = model.getIndexOfChild(parent,element);
	     for (int i=0; i<numChildren; i++) {
	  		child = (DefaultMutableTreeNode) childrenV.elementAt(i);
	  		model.insertNodeInto(child,parent,pos+1+i);
	    }
	 
		 e = getRoot().breadthFirstEnumeration();
         } 
       } // while
  } // removeNodes

 // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 // this function is called after a click in order to modify the tree w.r.t.
 // the new node state
 public void changedState(DefaultMutableTreeNode node, int preState, int postState, String comment) {

   if (preState == postState)
       return;

   reportAction(node,preState,postState,comment);

   // if the node is valid all its entailed nodes must be valid
   if (postState == NodeInfo.VALID )
       entailedValid(node);

    // if the node has become trusted all the nodes of the same function
    // must become trusted as well
    if (postState == NodeInfo.TRUSTED)
	    changeTrusted(node,true);

    // if the node has become nonvalid all those nodes that entail it are
    // also non valid
    if (postState == NodeInfo.NONVALID )
        entailsNonValid(node);

    // if a trusted node is not trusted, all the nodes corresponding to that
    // function are not trusted any more
    if (preState == NodeInfo.TRUSTED)
	    changeTrusted(node,false);

    updatePaint();

 } // changedState

void changeTrusted(DefaultMutableTreeNode node, boolean newValue) {
 // first extract the name of the function
 NodeInfo nodeInfo = (NodeInfo)(node.getUserObject());
 String functionName = nodeInfo.getFunctionName();
 int changed=0;

   Enumeration e = getRoot().breadthFirstEnumeration();
     while (e.hasMoreElements()) {
         DefaultMutableTreeNode element = (DefaultMutableTreeNode)e.nextElement();
         NodeInfo elementInfo = (NodeInfo)(element.getUserObject());
	 int elementState = elementInfo.getNodeState();
	 if (elementInfo.getFunctionName().equals(functionName))
	     if (newValue == true && elementState != NodeInfo.TRUSTED) {
	            elementInfo.setNodeState(NodeInfo.TRUSTED);
                     reportAction(element,elementState,NodeInfo.TRUSTED,"");
		     changed++;
		} else if (newValue == false && elementState == NodeInfo.TRUSTED) {
	                      elementInfo.setNodeState(NodeInfo.DONTKNOW);
                              reportAction(element,elementState,NodeInfo.DONTKNOW,"(no longer trusted)");
			      changed++;
		          }
     } // while
} // changeTrusted

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 void entailedValid(DefaultMutableTreeNode node) {
      NodeInfo nodeInfo = (NodeInfo)(node.getUserObject());
      Vector entailed = entailment.getEntails(node.toString());
      int changed=0;

 if (entailed.size() >0) {
    Enumeration e = getRoot().breadthFirstEnumeration();
     while (e.hasMoreElements()) {
         DefaultMutableTreeNode element = (DefaultMutableTreeNode)e.nextElement();
         NodeInfo elementInfo = (NodeInfo)(element.getUserObject());
	 for (int i=0; i<entailed.size(); i++) {
	      int preState = elementInfo.getNodeState();
	      if ( (entailed.get(i).equals(elementInfo.toString()) == true) &&
	            (preState != NodeInfo.VALID) ) {
		    elementInfo.setNodeState(NodeInfo.VALID);
                    reportAction(element,preState,NodeInfo.VALID,"(by entailment)");
		    changed++;
		}
          } // for
     } // while
    } // if
 } // entailedValid

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 void entailsNonValid(DefaultMutableTreeNode node) {
      NodeInfo nodeInfo = (NodeInfo)(node.getUserObject());
      Vector entailed = entailment.getEntailed(node.toString());
      int changed=0;

 if (entailed.size() >0) {
    Enumeration e = getRoot().breadthFirstEnumeration();
     while (e.hasMoreElements()) {
         DefaultMutableTreeNode element = (DefaultMutableTreeNode)e.nextElement();
         NodeInfo elementInfo = (NodeInfo)(element.getUserObject());
	 for (int i=0; i<entailed.size(); i++) {
	      int preState = elementInfo.getNodeState();
	      if ( (entailed.get(i).equals(elementInfo.toString()) == true) &&
	            (preState != NodeInfo.NONVALID) ) {
		    elementInfo.setNodeState(NodeInfo.NONVALID);
                    reportAction(element,preState,NodeInfo.NONVALID,"(by entailment)");
		    changed++;
		}
          } // for
     } // while
    } // if
 } // entailedNonValid

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 public void updatePaint() {
    calculate();
    repaint();

    if (getSelectionRows()== null)
             setSelectionRow(0);
 } // updatePaint


// show the parameter as an input node by means of a message dialog
void showBuggy(DefaultMutableTreeNode node) {
      NodeInfo nodeInfo = (NodeInfo)(node.getUserObject());

 	String msg = new String("Buggy Node Found!!!\n\n"+
						"Node: "+node.toString()+"\n"+
				                "Program rule: "+ nodeInfo.getFunctionName()+"."+
							   	               nodeInfo.getFunctionNumber()+"\n"+ "");
// We don't display the instance
//	 					"Rule instance: "+nodeInfo.getFunctionInstance());
       Toolkit.getDefaultToolkit().beep();
        ImageIcon bugIcon  = new ImageIcon(path+"/buggy.gif");

       JOptionPane.showMessageDialog(this,msg,"Buggy Node",
                                                           JOptionPane.PLAIN_MESSAGE,bugIcon);
 } // showBuggy

 boolean checkBuggy(DefaultMutableTreeNode node) {
     boolean isBuggy = false;
      NodeInfo nodeInfo =(NodeInfo)(node.getUserObject());

     if (nodeInfo.isRoot())
          return false;

      if (nodeInfo.getNodeState() == NodeInfo.NONVALID ||
           nodeInfo.getNodeState() == NodeInfo.BUGGY) {
        // check if all the children are valid; if so the node is buggy
        int childCount = model.getChildCount(node);
	boolean allValid = true;
	for (int i=0; allValid && i<childCount; i++) {
             DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node,i);
             NodeInfo childInfo =(NodeInfo)(child.getUserObject());
	     int childState = childInfo.getNodeState();
	     if (! (childState == NodeInfo.VALID || childState == NodeInfo.TRUSTED))
	           allValid = false;
	} // for

	// obtain the result
	if (allValid) {
            nodeInfo.setNodeState(NodeInfo.BUGGY);
	    isBuggy = true;
	 } else {
            nodeInfo.setNodeState(NodeInfo.NONVALID);
	    isBuggy = false;
	    }

      } // if
    return isBuggy;
 } // checkBuggy


 //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
} // ProofTree
