/*
 * Created on 24-sep-2005
 *
 * Represents a toy terms
 */
package terms;


import parserString.*;

import java.util.TreeSet;

/**
 * @author Rafael Caballero.
 *
 * Represents a Constraint Basic Fact
 */
public class BasicFact {
	private Term lhs;
	private Term rhs;
	private Condition condition;

	public BasicFact() {
		lhs=null;
		rhs=null;
		condition=null;
	}
		
	/**
	 * Constructs a new BasicFact from a String.
	 * @param s String to be parsed in order to create the BasicFact
	 */
	public BasicFact(String s) {
		String lhs,rhs,cond;
		
		int i1 = s.indexOf(" -> ");
		int i2 = s.indexOf(" <== ",i1+4);
		
		lhs  = s.substring(0,i1);

		this.lhs = TermParser.parse(lhs);
			
		// no conditions
		if (i2 == -1 ) {
			 rhs = s.substring(i1+4,s.length());
			 this.rhs = TermParser.parse(rhs);
			 condition = null;
		} else if (i2+4>=s.length()-1 ) {
			 rhs = s.substring(i1+4,s.length()-4);
			 this.rhs = TermParser.parse(rhs);
			 condition = null;			
		}
		else {
			rhs = s.substring(i1+4,i2);
			this.rhs = TermParser.parse(rhs);
			cond = s.substring(i2+4,s.length());
			condition = new Condition(TermParser.parseCondition(cond));
			
			// simplify the condition
			TreeSet treeLhs = this.lhs.variables();
			TreeSet treeRhs = this.rhs.variables();
			
			TreeSet tree=null;
			if (treeLhs!=null) {
	           tree = treeLhs;
			   if (treeRhs != null) 
				  tree.addAll(treeRhs);
			} else 
				tree = treeRhs;
			
			condition.simplify(tree);
		}
		
	   	   
	}

	public Term getLhs() {return lhs; }
	public void setLhs(Term lhs) {this.lhs = lhs; }
	public Term getRhs() {return rhs; }
	public void setRhs(Term rhs) {this.rhs = rhs; }
	public Condition getCondition() {return condition; }
	public void setCondition(Condition cond) {this.condition = cond; }

	
	public String toString(){
		String result = lhs + " -> " + rhs;
		
		if (condition != null && !condition.empty())
			result += " <== " + condition;
		
		return result;
	}  
}
